package application;

import java.awt.AWTException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import org.json.JSONException;
import org.opencv.core.Core;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.imgproc.Imgproc;
import org.opencv.videoio.VideoCapture;

import it.polito.elite.teaching.cv.utils.Utils;
import javafx.application.Platform;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.control.Slider;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.layout.BorderPane;
import javafx.stage.Stage;
import javafx.stage.StageStyle;
import speech.Voice;
import twitter4j.TwitterException;

/**
 * The controller associated with the only view of our application. The
 * application logic is implemented here. It handles the button for
 * starting/stopping the camera, the acquired video stream, the relative
 * controls and the image segmentation process.
 * 
 * @author <a href="mailto:luigi.derussis@polito.it">Luigi De Russis</a>
 * @version 2.0 (2017-03-10)
 * @since 1.0 (2015-01-13)
 * 
 */
public class SettingsController {
    // FXML button to switch over to the main functions.
    @FXML
    private Button Continue;

    // FXML button to change the setting to find blue
    @FXML
    private Button blueButton;
    // FXML button to change the setting to find green
    @FXML
    private Button greenButton;
    // FXML button to start speech recognition
    @FXML
    private Button speechRecognition;

    // FXML camera button
    @FXML
    private Button cameraButton;
    // the FXML area for showing the current frame
    // @FXML
    // private ImageView originalFrame;
    // the FXML area for showing the mask
    // @FXML
    // private ImageView maskImage;
    // the FXML area for showing the output of the morphological operations
    @FXML
    private ImageView morphImage;
    // FXML slider for setting HSV ranges
    @FXML
    private Slider hueStart;
    @FXML
    private Slider hueStop;
    @FXML
    private Slider saturationStart;
    @FXML
    private Slider saturationStop;
    @FXML
    private Slider valueStart;
    @FXML
    private Slider valueStop;
    // FXML label to show the current values set with the sliders
    @FXML
    private Label hsvCurrentValues;

    private int xLocation;
    private int yLocation;

    private List<Integer> leftRightMvmt = new ArrayList<>();
    private List<Integer> upDownMvmt = new ArrayList<>();

    private boolean left = false;
    private boolean right = false;
    private boolean up = false;
    private boolean down = false;

    // this is used to communicate this controller to HomeController
    private HomeController homeController;

    // a timer for acquiring the video stream
    private ScheduledExecutorService timer;
    // the OpenCV object that performs the video capture
    private VideoCapture capture = new VideoCapture();
    // a flag to change the button behavior
    private boolean cameraActive;

    // property for object binding
    private ObjectProperty<String> hsvValuesProp;

    // This will launch the actual program over the setting window
    @FXML
    void openHome(ActionEvent event) {

        try {
            Stage stage = new Stage();

            FXMLLoader loader = new FXMLLoader(
                    getClass().getResource("Home.fxml"));
            BorderPane root1 = (BorderPane) loader.load();
            stage.setTitle("Home screen");
            stage.setScene(new Scene(root1));
           // stage.setFullScreen(true);
            
            stage.setHeight(700);
            stage.setWidth(1400);
           // stage.centerOnScreen();
            
            stage.setX(110);
            stage.setY(98);
            stage.initStyle(StageStyle.UNDECORATED);
            
            stage.show();
            homeController = loader.getController();

        }
        catch (Exception e) {
            System.out.println("New Window will not load");
        }

        voiceCommands(event);
        startCamera();

    }

    // Action is triggered when pushing blue button
    @FXML
    void findBlue(ActionEvent event) {
        this.hueStart.setValue(70);
        this.hueStop.setValue(122);

        this.saturationStart.setValue(155);
        this.saturationStop.setValue(255);

        this.valueStart.setValue(20);
        this.valueStop.setValue(255);
    }

    @FXML
    void findGreen(ActionEvent event) {
        this.hueStart.setValue(40);
        this.hueStop.setValue(87);

        this.saturationStart.setValue(90);
        this.saturationStop.setValue(255);

        this.valueStart.setValue(20);
        this.valueStop.setValue(255);

    }

    /**
     * The action triggered by pushing the button on the GUI
     */
    @FXML
    private void startCamera() {

        // bind a text property with the string containing the current range of
        // HSV values for object detection
        hsvValuesProp = new SimpleObjectProperty<>();
        this.hsvCurrentValues.textProperty().bind(hsvValuesProp);

        // set a fixed width for all the image to show and preserve image ratio
        this.imageViewProperties(this.morphImage, 200);

        if (!this.cameraActive) {
            // start the video capture
            this.capture.open(0);

            // is the video stream available?
            if (this.capture.isOpened()) {
                this.cameraActive = true;

                // grab a frame every 33 ms (30 frames/sec)
                Runnable frameGrabber = new Runnable() {

                    @Override
                    public void run() {
                        // effectively grab and process a single frame
                        Mat frame = grabFrame();
                        // convert and show the frame
                        Image imageToShow = Utils.mat2Image(frame);
                        // updateImageView(originalFrame, imageToShow);
                    }
                };

                this.timer = Executors.newSingleThreadScheduledExecutor();
                this.timer.scheduleAtFixedRate(frameGrabber, 0, 33,
                        TimeUnit.MILLISECONDS);

                // update the button content
                this.cameraButton.setText("Stop Camera");
            }
            else {
                // log the error
                System.err.println("Failed to open the camera connection...");
            }
        }
        else {
            // the camera is not active at this point
            this.cameraActive = false;
            // update again the button content
            this.cameraButton.setText("Start Camera");

            // stop the timer
            this.stopAcquisition();
        }
    }

    /**
     * Get a frame from the opened video stream (if any)
     * 
     * @return the {@link Image} to show
     */
    private Mat grabFrame() {
        Mat frame = new Mat();

        // check if the capture is open
        if (this.capture.isOpened()) {
            try {
                // read the current frame
                this.capture.read(frame);

                // if the frame is not empty, process it
                if (!frame.empty()) {
                    // init
                    Mat blurredImage = new Mat();
                    Mat hsvImage = new Mat();
                    Mat mask = new Mat();
                    Mat morphOutput = new Mat();

                    // remove some noise
                    Imgproc.blur(frame, blurredImage, new Size(7, 7));

                    // convert the frame to HSV
                    Imgproc.cvtColor(blurredImage, hsvImage,
                            Imgproc.COLOR_BGR2HSV);

                    // get thresholding values from the UI
                    // remember: H ranges 0-180, S and V range 0-255
                    Scalar minValues = new Scalar(this.hueStart.getValue(),
                            this.saturationStart.getValue(),
                            this.valueStart.getValue());
                    Scalar maxValues = new Scalar(this.hueStop.getValue(),
                            this.saturationStop.getValue(),
                            this.valueStop.getValue());

                    // show the current selected HSV range
                    String valuesToPrint = "Hue range: " + minValues.val[0]
                            + "-" + maxValues.val[0] + "\tSaturation range: "
                            + minValues.val[1] + "-" + maxValues.val[1]
                            + "\tValue range: " + minValues.val[2] + "-"
                            + maxValues.val[2];
                    Utils.onFXThread(this.hsvValuesProp, valuesToPrint);

                    // threshold HSV image to select tennis balls
                    Core.inRange(hsvImage, minValues, maxValues, mask);
                    
                    // morphological operators
                    // dilate with large element, erode with small ones
                    Mat dilateElement = Imgproc.getStructuringElement(
                            Imgproc.MORPH_RECT, new Size(24, 24));
                    Mat erodeElement = Imgproc.getStructuringElement(
                            Imgproc.MORPH_RECT, new Size(12, 12));

                    Imgproc.erode(mask, morphOutput, erodeElement);
                    Imgproc.erode(morphOutput, morphOutput, erodeElement);

                    Imgproc.dilate(morphOutput, morphOutput, dilateElement);
                    Imgproc.dilate(morphOutput, morphOutput, dilateElement);

                    // show the partial output
                    this.updateImageView(this.morphImage,
                            Utils.mat2Image(morphOutput));

                    // find the tennis ball(s) contours and show them
                    frame = this.findAndDrawBalls(morphOutput, frame);

                }

            }
            catch (Exception e) {
                // log the (full) error
                System.err.print("Exception during the image elaboration...");
                e.printStackTrace();
            }
        }

        return frame;
    }

    /**
     * Given a binary image containing one or more closed surfaces, use it as a
     * mask to find and highlight the objects contours
     * 
     * @param maskedImage
     *            the binary image to be used as a mask
     * @param frame
     *            the original {@link Mat} image to be used for drawing the
     *            objects contours
     * @return the {@link Mat} image with the objects contours framed
     * @throws TwitterException
     * @throws JSONException
     * @throws IOException
     * @throws AWTException
     */
    private Mat findAndDrawBalls(Mat maskedImage, Mat frame)
            throws TwitterException, IOException, JSONException, AWTException {
        // init
        List<MatOfPoint> contours = new ArrayList<>();
        Mat hierarchy = new Mat();

        // find contours
        Imgproc.findContours(maskedImage, contours, hierarchy,
                Imgproc.RETR_CCOMP, Imgproc.CHAIN_APPROX_SIMPLE);

        // if any contour exist...
        if (hierarchy.size().height > 0 && hierarchy.size().width > 0) {

            // for each contour, display it in blue
            for (int idx = 0; idx >= 0; idx = (int) hierarchy.get(0, idx)[0]) {
                // creates a rectangle around contour
                Rect rect = Imgproc.boundingRect(contours.get(idx));
                Imgproc.rectangle(frame, rect.tl(), rect.br(),
                        new Scalar(250, 0, 0));

                xLocation = rect.x;
                yLocation = rect.y;
                leftRightMvmt.add(xLocation);
                upDownMvmt.add(yLocation);

                // used to detect movement
                if (leftRightMvmt.size() > 10) {
                    left = leftDownMvmt(leftRightMvmt);
                    right = rightUpMvmt(leftRightMvmt);

                    //
                    leftRightMvmt.clear();
                    if (left) {
                        // sends data to HomeController to control movement

                        homeController.getController().highlighted("left");

                    }
                    if (right) {
                        // sends data to HomeController to control movement
                        homeController.getController().highlighted("right");

                    }

                }

                if (upDownMvmt.size() > 15) {
                    up = rightUpMvmt(upDownMvmt);
                    down = leftDownMvmt(upDownMvmt);
                    upDownMvmt.clear();

                    if (up) {
                        homeController.getController().highlighted("up");

                    }
                    if (down) {
                        homeController.getController().highlighted("down");

                    }

                }

            }
        }

        return frame;
    }

    /**
     * Used to track negative number movement
     * 
     * @param xaxis
     *            direction in which the rectangle moves
     * @return boolean
     */
    private boolean leftDownMvmt(List<Integer> xaxis) {

        // Ascending
        for (int i = 0; i < xaxis.size() - 1; i++) {
            if (xaxis.get(i) > xaxis.get(i + 1)) {
                return false;
            }

        }
        return true;
    }

    /**
     * Used to track positive number movement
     * 
     * @param xaxis
     *            direction in which the rectangle moves
     * @return boolean
     */
    private boolean rightUpMvmt(List<Integer> xaxis) {
        // Descending
        for (int i = 0; i < xaxis.size() - 1; i++) {
            if (xaxis.get(i) < xaxis.get(i + 1)) {
                return false;
            }
        }
        return true;
    }

    /**
     * Set typical {@link ImageView} properties: a fixed width and the
     * information to preserve the original image ration
     * 
     * @param image
     *            the {@link ImageView} to use
     * @param dimension
     *            the width of the image to set
     */
    private void imageViewProperties(ImageView image, int dimension) {
        // set a fixed width for the given ImageView
        image.setFitWidth(dimension);
        // preserve the image ratio
        image.setPreserveRatio(true);
    }

    /**
     * Stop the acquisition from the camera and release all the resources
     */
    private void stopAcquisition() {
        if (this.timer != null && !this.timer.isShutdown()) {
            try {
                // stop the timer
                this.timer.shutdown();
                this.timer.awaitTermination(33, TimeUnit.MILLISECONDS);
            }
            catch (InterruptedException e) {
                // log any exception
                System.err.println(
                        "Exception in stopping the frame capture, trying to release the camera now... "
                                + e);
            }
        }

        if (this.capture.isOpened()) {
            // release the camera
            this.capture.release();
        }
    }

    /**
     * Update the {@link ImageView} in the JavaFX main thread
     * 
     * @param view
     *            the {@link ImageView} to update
     * @param image
     *            the {@link Image} to show
     */
    private void updateImageView(ImageView view, Image image) {
        Utils.onFXThread(view.imageProperty(), image);
    }

    /**
     * On application close, stop the acquisition from the camera
     */
    protected void setClosed() {
        this.stopAcquisition();
    }

    /**
     * sets HomeController object to voice
     * 
     * @param event
     *            on button click
     */
    @FXML
    private void voiceCommands(ActionEvent event) {

        Platform.runLater(() -> new Voice(homeController));
    }

}
